"""Custom Data Types for Sinai."""

from decimal import Decimal
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

from typing_extensions import TypeAlias

if TYPE_CHECKING:  # pragma: no cover
    from sinai.models.metric import AggregationMetric, Metric  # noqa: F401
    from sinai.models.monitor import Monitor  # noqa: F401
    from sinai.models.rule import Rule
    from sinai.models.source import MetricSource, Source  # noqa: F401
    from sinai.models.store import Store

    try:
        from bson.objectid import ObjectId  # pylint: disable=W0611
    except ImportError:
        pass


# Common Data Types
JDict = Dict[str, Any]  # A JSON-Safe Dictionary
JDictOrNone = Union[JDict, None]  # A JSON-Safe Dictionary or none
FList = List[Any]  # A Freeform List
SList = List[str]  # A List of Strings

# Metric types
MetricInstance: TypeAlias = "Metric"
MetricList = List[MetricInstance]
MetricClass = Type[MetricInstance]
MetricClasses = List[MetricClass]
MetricId = Union[UUID, "ObjectId", str]
MetricDict = Dict[MetricId, MetricInstance]
MetricResult = Tuple[MetricInstance, MetricId]
MetricValue = Union[Number, Decimal, None, int, float]
MetricValueList = List[MetricValue]
AggregationFunction = Callable[[MetricValueList], MetricValue]
AggregationMetrics = List["AggregationMetric"]
Evaluation = Union[List[MetricInstance], AggregationMetrics, MetricInstance, None]

# Monitor Types
MonitorInstance: TypeAlias = "Monitor"
GlobalMemory = Dict[MetricId, JDict]

# Rule types
RuleClass = Type["Rule"]
RuleClasses = List[RuleClass]

# Store types
StoreClass = Type["Store"]
StoreClassSet = Set[StoreClass]
StoreDict = Dict[StoreClass, "Store"]
StoreClasses = Sequence[StoreClass]

# Source types
SourceInstance: TypeAlias = "Source"
SourceClass = Type[SourceInstance]
SourceClassSet = Set[SourceClass]
SourceDict = Dict[SourceClass, SourceInstance]
SourceClasses = Sequence[SourceClass]

# MetricSource types
MetricSourceInstance: TypeAlias = "MetricSource"
MetricSourceClass = Type[MetricSourceInstance]
MetricSourceClasses = Sequence[MetricSourceClass]


# ApiSource types
RequestHeader = Dict[str, str]

# Mongo adapter types
MongoResult: TypeAlias = Any



"""
Monitor (gets all data, evaluates commandments
stores commandment outcome and stores metrics)
    get data
    run rule
        rules generate metrics (some rules are simple, some require processing)
    store metrics
"""
from collections.abc import Iterable
from uuid import uuid4

from sinai.models.metric import Metric
from sinai.models.rule import Rule
from sinai.types import (
    Evaluation,
    GlobalMemory,
    JDict,
    MetricDict,
    RuleClass,
    RuleClasses,
    SourceClass,
    SourceClassSet,
    SourceDict,
    SourceInstance,
    StoreClassSet,
    StoreDict,
)


class Monitor:
    """The monitor controls the monitoring run."""

    rules: RuleClasses = []

    def __init__(self) -> None:
        self.id = str(uuid4())
        self.sources: SourceClassSet = set()
        self.stores: StoreClassSet = set()
        self._store_instances: StoreDict = {}
        self._source_instances: SourceDict = {}
        self.memory: GlobalMemory = {}  # Used for in memory store/source
        self._resolve_rules()

    def _resolve_rules(self) -> None:
        for rule in self.rules:
            self.sources.update(rule.sources)
            self.stores.update(rule.stores)

    def _connect_sources(self) -> None:
        for source_class in self.sources:
            self._source_instances[source_class] = source_class(monitor=self)

    def _connect_stores(self) -> None:
        for store_class in self.stores:
            self._store_instances[store_class] = store_class(monitor=self)

    def _evaluate_rules(self) -> None:
        for rule_class in self.rules:
            self._evaluate_rule(rule_class)

    def _evaluate_rule(self, rule_class: RuleClass) -> None:
        rule = rule_class(self)
        result: Evaluation = rule.evaluate()
        if isinstance(result, Iterable):
            for metric in result:
                self.store_metric(rule, metric)
        elif not result:
            return
        else:
            self.store_metric(rule, result)

    def execute(self) -> None:
        """Start monitoring."""
        self._connect_sources()
        self._connect_stores()
        self._evaluate_rules()

    def source(self, source: SourceClass) -> SourceInstance:
        """Return an instatatied source."""
        return self._source_instances[source]

    def store_metric(self, rule: Rule, metric: Metric) -> None:
        """Store a metric."""
        for store_class in rule.stores:
            store = self._store_instances[store_class]
            store.save_metric(metric)

    def find_metric(self, metric_filter: JDict) -> MetricDict:
        """Find a metric in the in-memory store."""
        results = {}
        for metric_id, metric_dict in self.memory.items():
            for filter_key in metric_filter:
                if metric_dict[filter_key] != metric_filter[filter_key]:
                    continue
                metric = Metric.from_dict(metric_dict)
                results[metric_id] = metric
        return results



"""A Store is any output channel from a rule."""

from dataclasses import dataclass
from uuid import uuid4

from sinai.exceptions import MetricNotFound
from sinai.types import JDict, MetricId, MetricInstance, MetricResult, MonitorInstance


@dataclass
class Store:
    """Basic in memory metric store.
    This is the base class. You want to us an adaptors to get persistence."""

    def __init__(self, monitor: MonitorInstance) -> None:
        self.monitor = monitor

    def save_metric(self, metric: MetricInstance) -> None:
        """Save a metric instance into the store."""
        if metric.update:
            self._upsert(metric)
        else:
            self._save(metric)

    def _save(self, metric: MetricInstance) -> None:
        self.monitor.memory[uuid4()] = metric.pre_save(self.monitor.id)

    def _upsert(self, metric: MetricInstance) -> None:
        try:
            existing, metric_id = self._get_metric_for_update(metric)
        except MetricNotFound:
            self._save(metric)
        else:
            self._replace(existing, metric_id)

    def _replace(self, metric: MetricInstance, metric_id: MetricId) -> None:
        self.monitor.memory[metric_id] = metric.pre_save(self.monitor.id)

    def _get_metric_for_update(self, metric: MetricInstance) -> MetricResult:
        metric_filter: JDict = {}
        for field in metric.update:
            metric_filter[field] = getattr(metric, field)
        return self._find_metric_by_filter(metric, metric_filter)

    def _find_metric_by_filter(
        self, metric: MetricInstance, metric_filter: JDict
    ) -> MetricResult:
        if results := self.monitor.find_metric(metric_filter):
            for metric_id, existing_metric in results.items():
                metric.created_at = existing_metric.created_at
                # This could support updating multiple metrics, for now support only the first
                return metric, metric_id
        raise MetricNotFound(f"Metric with fields: {metric_filter} not found.")



"""Support for using Mongo Databases as a source or store for data."""

from __future__ import annotations

import pymongo

from sinai.exceptions import MetricNotFound
from sinai.models.metric import Metric
from sinai.models.source import MetricSource, Source
from sinai.models.store import Store
from sinai.types import (
    FList,
    JDict,
    JDictOrNone,
    MetricId,
    MetricInstance,
    MetricList,
    MetricResult,
    MongoResult,
    MonitorInstance,
)


class MongoConnection:
    """A connection to MongoDB."""

    connection_string = ""
    database_name = ""

    def __init__(self, monitor: MonitorInstance) -> None:
        self.monitor = monitor
        self.client = pymongo.MongoClient(self.connection_string)  # type: ignore
        self.db = self.client[self.database_name]
        self.collection = self.db["metrics"]


class MongoStore(MongoConnection, Store):
    """Metric store using MongoDB. Don't forget to set `database_name` and `connection_string`."""

    def _save(self, metric: MetricInstance) -> None:
        self.collection.insert_one(metric.pre_save(self.monitor.id))

    def _find_metric_by_filter(
        self, metric: MetricInstance, metric_filter: JDict
    ) -> MetricResult:
        if result := self.collection.find_one(metric_filter):
            mongo_id = result.pop("_id")
            existing_metric = Metric.from_dict(result)
            metric.created_at = existing_metric.created_at
            return metric, mongo_id
        raise MetricNotFound(f"Metric with fields: {metric_filter} not found.")

    def _replace(self, metric: MetricInstance, metric_id: MetricId) -> None:
        self.collection.replace_one(
            {"_id": metric_id}, metric.pre_save(self.monitor.id)
        )


class MongoSource(MongoConnection, Source):
    """Gets arbitrary documents from MongoDB."""

    def find_one(
        self,
        collection: str,
        mongo_filter: JDictOrNone = None,
        *args: FList,
        **kwargs: JDict,
    ) -> MongoResult:
        """Find a single document from MongoFB."""
        return self.db[collection].find_one(mongo_filter, *args, **kwargs)

    def find(
        self,
        collection: str,
        mongo_filter: JDictOrNone = None,
        *args: FList,
        **kwargs: JDict,
    ) -> MongoResult:
        """Find documents from MongoFB."""
        return self.db[collection].find(mongo_filter, *args, **kwargs)


class MongoMetricSource(MongoSource, MetricSource):
    """Gets metrics from MongoDB."""

    def _execute_query(self, metric_filter: JDict) -> MetricList:
        return [
            self._document_to_metric(document)
            for document in self.collection.find(metric_filter)
        ]

    @staticmethod
    def _document_to_metric(document: JDict) -> MetricInstance:
        del document["_id"]
        return Metric.from_dict(document)



"""The Api adaptors provide sources and stores using API endpoints.
Requires the requests library to be installed.
"""
try:
    import requests  # type: ignore
except ImportError:
    HAS_REQUESTS = False
else:
    HAS_REQUESTS = True

from sinai.exceptions import SourceError
from sinai.models.source import Source
from sinai.models.store import Store
from sinai.types import JDict, MonitorInstance, RequestHeader


class ApiConnection:
    """Base class to share common connection information between API Stores and Sources"""

    ...


class ApiStore(Store):
    """Store data using an API (not implemented)"""

    ...


class ApiSource(Source):
    """Get data from an API endpoint."""

    url: str = ""
    content: JDict = {}
    headers: RequestHeader = {}
    token: str = ""

    def __init__(self, monitor: MonitorInstance):
        super().__init__(monitor)
        if not HAS_REQUESTS:
            raise SourceError("Requests library required for API calls.")
        self.get()

    def get_headers(self) -> RequestHeader:
        """The required authentication, by default uses a bearer token."""
        if self.token:
            return self.bearer_headers(self.token)
        return self.headers

    @staticmethod
    def bearer_headers(token: str) -> RequestHeader:
        """Headers for APIs with bearer token authentication."""
        return {
            "accept": "application/json",
            "Authorization": "Bearer " + token,
        }

    def get(self) -> None:
        """Get the required data from the specified API."""
        response = requests.get(self.url, headers=self.get_headers())
        if response.status_code == 200:
            self.content = response.json()
        else:
            raise SourceError(
                f"The URL {self.url} returned status {response.status_code}."
            )



"""A Metric is the output of a rule, and often the input to rules too.
Metrics can be stored in a Store, and can be retrieved by a MetricSource."""
import datetime
from statistics import mean, median, mode
from uuid import uuid4

from sinai.types import AggregationFunction, JDict, MetricValue, MetricValueList, SList


class Metric:
    """A measure to be observed."""

    name: str = "monitoring_metric"
    context: str = ""
    update: SList = []

    def __init__(
        self,
        value: MetricValue = None,
        name: str = "",
        ref: str = "",
        context: str = "",
        update: SList = [],
    ) -> None:
        if name:
            self.name = name
        if context:
            self.context = context
        if update:
            self.update = update
        self.annotations: SList = []
        self.monitor_id: str = ""
        self.ref = ref or str(uuid4())
        self.value = value
        self.created_at = datetime.datetime.now(tz=datetime.timezone.utc)
        self.updated_at = self.created_at

    def pre_save(self, monitor_id: str) -> JDict:
        """Readythe Metric for storage."""
        self.updated_at = datetime.datetime.now(tz=datetime.timezone.utc)
        self.monitor_id = monitor_id
        return self.to_dict()

    def to_dict(self) -> JDict:
        """Serialise the Metric to a dictionary."""
        return {
            "name": self.name,
            "ref": self.ref,
            "value": self.value,
            "context": self.context,
            "annotations": self.annotations,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "monitor_id": self.monitor_id,
        }

    @classmethod
    def from_dict(cls, metric_dict: JDict) -> "Metric":
        """Deserialise a dictionary back to a Metric."""
        metric = cls(
            name=metric_dict["name"],
            ref=metric_dict["ref"],
            value=metric_dict["value"],
            context=metric_dict["context"],
        )
        metric.created_at = metric_dict["created_at"]
        metric.updated_at = metric_dict["updated_at"]
        metric.annotations = metric_dict["annotations"]
        metric.monitor_id = metric_dict["monitor_id"]
        return metric

    def annotate(self, text: str) -> None:
        """Store a piece of text in the Metric."""
        self.annotations.append(text)


class Sinner(Metric):
    """An unexpected or missing item between two sequences."""

    ...


class AggregationMetric(Metric):
    """A metric formed from summarizing a query of stored metrics."""

    name = "aggregation"
    aggregation: AggregationFunction = len

    def process_data_points(self, data_points: MetricValueList) -> None:
        """Turn multiple values into one."""
        self.value = self.aggregation(data_points)  # type:ignore


class CountMetric(AggregationMetric):
    """The total number of matching metrics."""

    name = "count"
    aggregation = len


class SumMetric(AggregationMetric):
    """The total values of matching metric."""

    name = "sum"
    aggregation = sum  # type:ignore


class MeanMetric(AggregationMetric):
    """The average of matching metric values."""

    name = "mean"
    aggregation = mean


class MaxMetric(AggregationMetric):
    """The maximum of the matching metric values."""

    name = "max"
    aggregation = max  # type:ignore


class MinMetric(AggregationMetric):
    """The minimum of the matching metric values."""

    name = "min"
    aggregation = min  # type:ignore


class MedianMetric(AggregationMetric):
    """The median of the matching metric values."""

    name = "median"
    aggregation = median


class ModeMetric(AggregationMetric):
    """The mode of the matching metric values."""

    name = "mode"
    aggregation = mode


AGGREGATION_CLASSES = {
    "count": CountMetric,
    "sum": SumMetric,
    "mean": MeanMetric,
    "max": MaxMetric,
    "min": MinMetric,
    "median": MedianMetric,
    "mode": ModeMetric,
}



"""Exceptions for Sinai."""


class MetricNotFound(Exception):
    """The search did not yield any result."""

    ...


class SourceError(Exception):
    """The Source did not return the required result."""

    ...



"""A `Source` provides the input data to the Rule."""

from sinai.types import JDict, MetricList, MonitorInstance


class Source:
    """Any data source (base class)."""

    def __init__(self, monitor: MonitorInstance):
        self.monitor = monitor


class MetricSource(Source):
    """In memory source of Metrics, use an adaptor instead when you need persistence."""

    def get(
        self,
        name: str = "",
        ref: str = "",
        value: str = "",
        context: str = "",
        monitor_id: str = "",
    ) -> MetricList:
        """Get a metric/metrics from the Source"""
        metric_filter = {}
        if name:
            metric_filter["name"] = name
        if ref:
            metric_filter["ref"] = ref
        if value:
            metric_filter["value"] = value
        if context:
            metric_filter["context"] = context
        if monitor_id:
            metric_filter["monitor_id"] = monitor_id
        return self._execute_query(metric_filter)

    def _execute_query(self, metric_filter: JDict) -> MetricList:
        results = self.monitor.find_metric(metric_filter)
        return list(results.values())
